@@ -53,13 +53,15 @@
 #include "stun/usages/timer.h"
 #include "agent-priv.h"
 
+#define STUN_END_TIMEOUT 8000
+
+
 typedef struct {
   StunMessage message;
   uint8_t buffer[STUN_MAX_MESSAGE_SIZE];
   StunTimer timer;
 } TURNMessage;
 
-
 typedef struct {
   NiceAddress peer;
   uint16_t channel;
@@ -80,9 +82,16 @@ typedef struct {
   uint8_t *password;
   size_t password_len;
   NiceTurnSocketCompatibility compatibility;
+  GQueue *send_requests;
 } TurnPriv;
 
 
+typedef struct {
+  StunTransactionId id;
+  GSource *source;
+  TurnPriv *priv;
+  gint ref;
+} SendRequest;
 
 static void socket_close (NiceSocket *sock);
 static gint socket_recv (NiceSocket *sock, NiceAddress *from,
@@ -99,7 +108,7 @@ static void priv_send_turn_message (TurnPriv *priv, TURNMessage *msg);
 static gboolean priv_send_channel_bind (TurnPriv *priv,  StunMessage *resp,
     uint16_t channel, NiceAddress *peer);
 static gboolean priv_add_channel_binding (TurnPriv *priv, NiceAddress *peer);
-
+static gboolean priv_forget_send_request (gpointer pointer);
 
 
 NiceSocket *
@@ -151,6 +160,7 @@ nice_turn_socket_new (NiceAgent *agent, NiceAddress *addr,
   }
   priv->server_addr = *server_addr;
   priv->compatibility = compatibility;
+  priv->send_requests = g_queue_new ();
   sock->addr = *addr;
   sock->fileno = base_socket->fileno;
   sock->send = socket_send;
@@ -167,6 +177,7 @@ socket_close (NiceSocket *sock)
 {
   TurnPriv *priv = (TurnPriv *) sock->priv;
   GList *i = NULL;
+
   for (i = priv->channels; i; i = i->next) {
     ChannelBinding *b = i->data;
     g_free (b);
@@ -185,6 +196,15 @@ socket_close (NiceSocket *sock)
     priv->tick_source = NULL;
   }
 
+  for (i = g_queue_peek_head_link (priv->send_requests); i; i = i->next) {
+    SendRequest *r = i->data;
+    g_source_destroy (r->source);
+    g_source_unref (r->source);
+    g_slice_free (SendRequest, r);
+  }
+  g_queue_free (priv->send_requests);
+
+
   g_free (priv->current_binding);
   g_free (priv->current_binding_msg);
   g_free (priv->username);
@@ -204,8 +224,11 @@ socket_recv (NiceSocket *sock, NiceAddress *from, guint len, gchar *buf)
   recv_len = nice_socket_recv (priv->base_socket, &recv_from,
       sizeof(recv_buf), (gchar *) recv_buf);
 
-  return nice_turn_socket_parse_recv (sock, &dummy, from, len, buf,
-      &recv_from, (gchar *) recv_buf, (guint) recv_len);
+  if (recv_len > 0)
+    return nice_turn_socket_parse_recv (sock, &dummy, from, len, buf,
+        &recv_from, (gchar *) recv_buf, (guint) recv_len);
+  else
+    return recv_len;
 }
 
 static gboolean
@@ -231,14 +254,17 @@ socket_send (NiceSocket *sock, const NiceAddress *to,
   nice_address_copy_to_sockaddr (to, (struct sockaddr *)&sa);
 
   if (binding) {
-    if (priv->compatibility == NICE_TURN_SOCKET_COMPATIBILITY_DRAFT9 &&
-        len + sizeof(uint32_t) <= sizeof(buffer)) {
-      uint16_t len16 = htons ((uint16_t) len);
-      uint16_t channel16 = htons (binding->channel);
-      memcpy (buffer, &channel16, sizeof(uint16_t));
-      memcpy (buffer + sizeof(uint16_t), &len16,sizeof(uint16_t));
-      memcpy (buffer + sizeof(uint32_t), buf, len);
-      msg_len = len + sizeof(uint32_t);
+    if (priv->compatibility == NICE_TURN_SOCKET_COMPATIBILITY_DRAFT9) {
+      if (len + sizeof(uint32_t) <= sizeof(buffer)) {
+        uint16_t len16 = htons ((uint16_t) len);
+        uint16_t channel16 = htons (binding->channel);
+        memcpy (buffer, &channel16, sizeof(uint16_t));
+        memcpy (buffer + sizeof(uint16_t), &len16,sizeof(uint16_t));
+        memcpy (buffer + sizeof(uint32_t), buf, len);
+        msg_len = len + sizeof(uint32_t);
+      } else {
+        return 0;
+      }
     } else {
       return nice_socket_send (priv->base_socket, &priv->server_addr, len, buf);
     }
@@ -282,6 +308,16 @@ socket_send (NiceSocket *sock, const NiceAddress *to,
 
     msg_len = stun_agent_finish_message (&priv->agent, &msg,
         priv->password, priv->password_len);
+    if (msg_len > 0 && stun_message_get_class (&msg) == STUN_REQUEST) {
+      SendRequest *req = g_slice_new0 (SendRequest);
+
+      req->priv = priv;
+      stun_message_id (&msg, req->id);
+      req->source = agent_timeout_add_with_context (priv->nice, STUN_END_TIMEOUT,
+          priv_forget_send_request, req);
+      g_queue_push_tail (priv->send_requests, req);
+      g_atomic_int_inc (&req->ref);
+    }
   }
 
   if (msg_len > 0) {
@@ -299,6 +335,32 @@ socket_is_reliable (NiceSocket *sock)
   return nice_socket_is_reliable (priv->base_socket);
 }
 
+static gboolean
+priv_forget_send_request (gpointer pointer)
+{
+  SendRequest *req = pointer;
+
+  g_atomic_int_inc (&req->ref);
+
+  g_static_rec_mutex_lock (&req->priv->nice->mutex);
+
+  stun_agent_forget_transaction (&req->priv->agent, req->id);
+
+  g_source_destroy (req->source);
+  g_source_unref (req->source);
+
+  if (g_queue_index (req->priv->send_requests, req) != -1) {
+    g_queue_remove (req->priv->send_requests, req);
+    (void)g_atomic_int_dec_and_test (&req->ref);
+  }
+
+  g_static_rec_mutex_unlock (&req->priv->nice->mutex);
+
+  if (g_atomic_int_dec_and_test (&req->ref))
+      g_slice_free (SendRequest, req);
+
+  return FALSE;
+}
 
 
 gint
@@ -330,12 +392,37 @@ nice_turn_socket_parse_recv (NiceSocket *sock, NiceSocket **from_sock,
       }
 
       if (stun_message_get_method (&msg) == STUN_SEND) {
-        if (stun_message_get_class (&msg) == STUN_RESPONSE &&
-            priv->compatibility == NICE_TURN_SOCKET_COMPATIBILITY_GOOGLE) {
-          uint32_t opts = 0;
-          if (stun_message_find32 (&msg, STUN_ATTRIBUTE_OPTIONS, &opts) ==
-              STUN_MESSAGE_RETURN_SUCCESS && opts & 0x1)
-            goto msn_google_lock;
+        if (stun_message_get_class (&msg) == STUN_RESPONSE) {
+          SendRequest *req = NULL;
+          GList *i = g_queue_peek_head_link (priv->send_requests);
+          StunTransactionId msg_id;
+
+          stun_message_id (&msg, msg_id);
+
+          for (; i; i = i->next) {
+            SendRequest *r = i->data;
+            if (memcmp (&r->id, msg_id, sizeof(StunTransactionId)) == 0) {
+              req = r;
+              break;
+            }
+          }
+
+          if (req) {
+            g_source_destroy (req->source);
+            g_source_unref (req->source);
+
+            g_queue_remove (priv->send_requests, req);
+
+            if (g_atomic_int_dec_and_test (&req->ref))
+              g_slice_free (SendRequest, req);
+          }
+
+          if (priv->compatibility == NICE_TURN_SOCKET_COMPATIBILITY_GOOGLE) {
+            uint32_t opts = 0;
+            if (stun_message_find32 (&msg, STUN_ATTRIBUTE_OPTIONS, &opts) ==
+                STUN_MESSAGE_RETURN_SUCCESS && opts & 0x1)
+              goto msn_google_lock;
+          }
         }
         return 0;
       } else if (stun_message_get_method (&msg) == STUN_OLD_SET_ACTIVE_DST) {
@@ -517,13 +604,21 @@ priv_retransmissions_tick_unlocked (TurnPriv *priv)
   if (priv->current_binding_msg) {
     switch (stun_timer_refresh (&priv->current_binding_msg->timer)) {
       case STUN_USAGE_TIMER_RETURN_TIMEOUT:
-        /* Time out */
-        g_free (priv->current_binding);
-        priv->current_binding = NULL;
-        g_free (priv->current_binding_msg);
-        priv->current_binding_msg = NULL;
-        priv_process_pending_bindings (priv);
-        break;
+        {
+          /* Time out */
+          StunTransactionId id;
+
+          g_free (priv->current_binding);
+          priv->current_binding = NULL;
+          g_free (priv->current_binding_msg);
+          priv->current_binding_msg = NULL;
+
+          stun_message_id (&priv->current_binding_msg->message, id);
+          stun_agent_forget_transaction (&priv->agent, id);
+
+          priv_process_pending_bindings (priv);
+          break;
+        }
       case STUN_USAGE_TIMER_RETURN_RETRANSMIT:
         /* Retransmit */
         nice_socket_send (priv->base_socket, &priv->server_addr,
